// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::collections::{BTreeMap, HashMap};

use dcolor::{Color, Rgb};
use dimage::{self, ImageExt, VecImage2};
use dvec::{Rect2, Vec2};
use try_from::TryFrom;

use ::Coord;
use error::{Error, Result};
use text::FallbackValue;

/// Terminal font.
pub struct Font<C> {
    glyph_size: Vec2<Coord>,
    glyphs: Vec<VecImage2<Rgb<u8>>>,
    char_map: BTreeMap<C, usize>,
}

impl<C: Ord> Font<C> {
    pub fn char_map (&self) -> &BTreeMap<C, usize> { &self.char_map }
    pub fn glyph_size (&self) -> Vec2<Coord> { self.glyph_size }
    pub fn glyphs (&self) -> &[VecImage2<Rgb<u8>>] { self.glyphs.as_ref() }

    pub fn init<T> (src: T) -> Result<Font<C>>
        where T: IntoIterator<Item = (C, VecImage2<Rgb<u8>>)>
    {
        let mut glyph_size = None;
        let mut glyphs = Vec::new();
        let mut glyph_map = HashMap::new();
        let mut char_map = BTreeMap::new();

        for (ch, glyph) in src.into_iter() {
            let size = glyph.size();

            if let Some(expected_size) = glyph_size {
                if size != expected_size {
                    return Err(Error::InvalidGlyphSize);
                }
            } else if size.x < 1 || size.y < 1 {
                return Err(Error::InvalidGlyphSize);
            } else {
                glyph_size = Some(size);
            }

            let glyph_index = match glyph_map.get(&glyph) {
                None => glyphs.len(),
                Some(&index) => index,
            };

            if glyph_index == glyphs.len() {
                glyphs.push(glyph.clone());
                glyph_map.insert(glyph, glyph_index);
            }

            if let Some(_) = char_map.insert(ch, glyph_index) {
                return Err(Error::DuplicateChar);
            }
        }

        match glyph_size {
            None => Err(Error::NoChars),
            Some(glyph_size) => Ok(Font { glyph_size, glyphs, char_map }),
        }
    }

    #[cfg(feature = "built-in-font")]
    pub fn init_built_in () -> Font<C>
        where ::cp437::Char: ::try_from::TryInto<C>
    {
        Font::init(::built_in_font::Loader::new()).expect("failed to load built-in font")
    }
}

/// OpenGL-compatible texture for a font.
#[cfg_attr(not(feature = "gl"), allow(dead_code))]
pub struct UvFont<C> {
    image: VecImage2<Rgb<u8>>,
    char_map: BTreeMap<C, Rect2<f32>>,
    fallback: Rect2<f32>,
    glyph_size: Vec2<Coord>,
}

impl<C: Ord> UvFont<C> {
    pub fn empty () -> UvFont<C> {
        UvFont {
            image: VecImage2::empty(),
            char_map: BTreeMap::new(),
            fallback: Rect2(Vec2::new(0.0, 0.0), Vec2::new(0.0, 0.0)),
            glyph_size: Vec2::new(0, 0),
        }
    }

    pub fn get (&self, ch: C) -> Rect2<f32> {
        match self.char_map.get(&ch) {
            None => self.fallback,
            Some(&uv) => uv,
        }
    }

    pub fn glyph_size (&self) -> Vec2<Coord> { self.glyph_size }
    pub fn image (&self) -> &VecImage2<Rgb<u8>> { &self.image }

    pub fn init (font: &Font<C>) -> Result<UvFont<C>>
        where C: Clone + Eq + FallbackValue
    {
        let fallback_char = C::fallback_value();
        let glyph_size = font.glyph_size;

        // Compute image layout
        let mut dimension = 128;
        let mut image_size;
        let mut layout;

        loop {
            image_size = Vec2::new(dimension, dimension);
            layout = image_size / glyph_size;

            let max_glyphs = (image_size / glyph_size).product() as usize;

            if max_glyphs >= font.glyphs.len() {
                break;
            } else {
                dimension *= 2;
            }
        }

        // Build image
        let mut image = dimage::fill(Rgb::black()).with_size2(image_size).render_vec2();

        for i in 0..font.glyphs.len() {
            let pos = glyph_size * (i as Coord % layout.x, i as Coord / layout.x);
            image.blit(pos, &font.glyphs[i]);
        }

        // Build UV map
        let f_image_size = image_size.raw_conv::<f32>();
        let mut char_map = BTreeMap::new();
        let mut fallback_uv = None;

        for (ch, &glyph_index) in font.char_map.iter() {
            let i = Coord::try_from(glyph_index).unwrap();
            let pos = glyph_size * (i as Coord % layout.x, i as Coord / layout.x);
            let rect = pos.with_size(glyph_size);
            let uv = rect.raw_conv::<f32>() / f_image_size;
            char_map.insert(ch.clone(), uv);

            if *ch == fallback_char {
                fallback_uv = Some(uv);
            }
        }

        let fallback = match fallback_uv {
            None => { return Err(Error::NoFallbackGlyph); },
            Some(f) => f,
        };

        Ok(UvFont { image, char_map, fallback, glyph_size })
    }
}
