// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::error::Error;
use std::fmt::{self, Display, Formatter};
use std::mem;

use try_from::TryFrom;

use cp437;
use text::FallbackValue;

const CHAR_ERROR_STR: &'static str = "not a printable ascii char";

/// Printable ASCII character.
#[derive(Clone, Copy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
#[repr(u8)]
pub enum Char {
    Space = 0x20,
    Exclamation = 0x21,
    Quotation = 0x22,
    Number = 0x23,
    Dollar = 0x24,
    Percent = 0x25,
    Ampersand = 0x26,
    Apostrophe = 0x27,
    LParenthesis = 0x28,
    RParenthesis = 0x29,
    Asterisk = 0x2a,
    Plus = 0x2b,
    Comma = 0x2c,
    Minus = 0x2d,
    Dot = 0x2e,
    Slash = 0x2f,
    Digit0 = 0x30,
    Digit1 = 0x31,
    Digit2 = 0x32,
    Digit3 = 0x33,
    Digit4 = 0x34,
    Digit5 = 0x35,
    Digit6 = 0x36,
    Digit7 = 0x37,
    Digit8 = 0x38,
    Digit9 = 0x39,
    Colon = 0x3a,
    Semicolon = 0x3b,
    Less = 0x3c,
    Equal = 0x3d,
    Greater = 0x3e,
    Question = 0x3f,
    At = 0x40,
    UpperA = 0x41,
    UpperB = 0x42,
    UpperC = 0x43,
    UpperD = 0x44,
    UpperE = 0x45,
    UpperF = 0x46,
    UpperG = 0x47,
    UpperH = 0x48,
    UpperI = 0x49,
    UpperJ = 0x4a,
    UpperK = 0x4b,
    UpperL = 0x4c,
    UpperM = 0x4d,
    UpperN = 0x4e,
    UpperO = 0x4f,
    UpperP = 0x50,
    UpperQ = 0x51,
    UpperR = 0x52,
    UpperS = 0x53,
    UpperT = 0x54,
    UpperU = 0x55,
    UpperV = 0x56,
    UpperW = 0x57,
    UpperX = 0x58,
    UpperY = 0x59,
    UpperZ = 0x5a,
    LBracket = 0x5b,
    Backslash = 0x5c,
    RBracket = 0x5d,
    Caret = 0x5e,
    Underscore = 0x5f,
    Grave = 0x60,
    LowerA = 0x61,
    LowerB = 0x62,
    LowerC = 0x63,
    LowerD = 0x64,
    LowerE = 0x65,
    LowerF = 0x66,
    LowerG = 0x67,
    LowerH = 0x68,
    LowerI = 0x69,
    LowerJ = 0x6a,
    LowerK = 0x6b,
    LowerL = 0x6c,
    LowerM = 0x6d,
    LowerN = 0x6e,
    LowerO = 0x6f,
    LowerP = 0x70,
    LowerQ = 0x71,
    LowerR = 0x72,
    LowerS = 0x73,
    LowerT = 0x74,
    LowerU = 0x75,
    LowerV = 0x76,
    LowerW = 0x77,
    LowerX = 0x78,
    LowerY = 0x79,
    LowerZ = 0x7a,
    LBrace = 0x7b,
    Bar = 0x7c,
    RBrace = 0x7d,
    Tilde = 0x7e,
}

impl Char {
    pub fn as_byte (self) -> u8 { self as u8 }
    pub fn as_ucs (self) -> char { self as u8 as char }
}

impl FallbackValue for Char {
    fn fallback_value () -> Char { Char::Question }
}

impl Into<u8> for Char {
    fn into (self) -> u8 { self as u8 }
}

impl Into<char> for Char {
    fn into (self) -> char { self as u8 as char }
}

impl TryFrom<u8> for Char {
    type Err = CharError;

    fn try_from (byte: u8) -> Result<Char, CharError> {
        match byte {
            0x20...0x7e => {
                unsafe {
                    Ok(mem::transmute(byte))
                }
            },
            _ => Err(CharError)
        }
    }
}

impl TryFrom<char> for Char {
    type Err = CharError;

    fn try_from (ch: char) -> Result<Char, CharError> {
        match u8::try_from(ch) {
            Ok(byte) => Char::try_from(byte),
            Err(_) => Err(CharError),
        }
    }
}

impl TryFrom<cp437::Char> for Char {
    type Err = CharError;

    fn try_from (cp437: cp437::Char) -> Result<Char, CharError> {
        Char::try_from(cp437.as_byte())
    }
}

/// Error which occurs when an invalid `pascii` char is encountered.
#[derive(Clone, Copy, Debug)]
pub struct CharError;

impl Display for CharError {
    fn fmt (&self, f: &mut Formatter) -> fmt::Result {
        f.write_str(CHAR_ERROR_STR)
    }
}

impl Error for CharError {
    fn description (&self) -> &str { CHAR_ERROR_STR }
}
