// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::mem;
use std::rc::Rc;

use dcolor::{Color, Rgb, Rgba};
use dgl::{Config, ConfigInfo, Context, DisplayExt, PixelFormat, RgbFormat, Version};
use dvec::{Rect2, Vec2};
use dwindow::{Display, Window};

use dterm_gl::gl21 as gl;
use dterm_gl::gl21::types::*;

use ::Coord;
use error::{Error, Result};
use font::{Font, UvFont};
use render::Render;
use terminal::{Gravity, TerminalBuilder};
use text::FallbackValue;

/// Vertex attributes.
#[derive(Clone, Copy)]
struct Attrib {
    vertex: Vec2<GLint>,
    tex_coord: Vec2<GLfloat>,
    color: Rgba<GLubyte>,
}

type Quad = [Attrib; 4];

/// OpenGL 2.1 renderer.
pub struct Renderer<C> {
    context: Context,
    gl: gl::Gl,
    font: UvFont<C>,
    font_tex: GLuint,
    gravity: Gravity,
    scale: Vec2<Coord>,
    buf_size: Vec2<Coord>,
    bg_buf: Vec<Quad>,
    fg_buf: Vec<Quad>,
    dirty: bool,
}

impl<C> Renderer<C>
    where C: Copy + FallbackValue + Ord
{
    fn init_vertices (&mut self) {
        let cell_size = self.font.glyph_size() * self.scale;
        let mut i = 0;

        for y in 0..self.buf_size.y {
            for x in 0..self.buf_size.x {
                let pos = cell_size * (x, y);

                self.bg_buf[i][0].vertex = pos;
                self.bg_buf[i][1].vertex = pos + (0, cell_size.y);
                self.bg_buf[i][2].vertex = pos + cell_size;
                self.bg_buf[i][3].vertex = pos + (cell_size.x, 0);

                self.fg_buf[i][0].vertex = pos;
                self.fg_buf[i][1].vertex = pos + (0, cell_size.y);
                self.fg_buf[i][2].vertex = pos + cell_size;
                self.fg_buf[i][3].vertex = pos + (cell_size.x, 0);

                i += 1;
            }
        }
    }
}

impl<C> Render<C> for Renderer<C>
    where C: Copy + FallbackValue + Ord
{
    type Visual = Config;

    fn clear (&mut self, color: Rgb<u8>) {
        let bg = color.with_alpha(0xff);
        let fg = Rgba::new(0, 0, 0, 0);

        for cell in self.bg_buf.iter_mut() {
            cell[0].color = bg;
            cell[1].color = bg;
            cell[2].color = bg;
            cell[3].color = bg;
        }

        for cell in self.fg_buf.iter_mut() {
            cell[0].color = fg;
            cell[1].color = fg;
            cell[2].color = fg;
            cell[3].color = fg;
        }
    }

    fn choose_visual (display: &Display) -> Result<Config> {
        Ok(display.choose_gl_config(&ConfigInfo {
            draw_to_window: true,
            pixel_format: PixelFormat::Rgb(RgbFormat {
                red_size: 8,
                green_size: 8,
                blue_size: 8,
                ..RgbFormat::default()
            }),
            double_buffer: true,
            ..ConfigInfo::default()
        })?)
    }

    fn damage (&mut self, _: Rect2<Coord>) {
        self.dirty = true;
    }

    fn init (builder: &TerminalBuilder<C>, window: &Window, config: Config) -> Result<Renderer<C>> {
        let display = builder.display();
        let font = builder.font().unwrap();

        // Create OpenGL context.
        info!("Requesting OpenGL version 2.1...");

        let context = display.new_gl_context()
                             .with_config(&config)
                             .with_version((2, 1))
                             .build()?;

        // Initialize OpenGL state.
        let gl = context.with(window, |glc| {
            match glc.version() {
                Ok(version) => {
                    if version >= Version(2, 1) && version <= Version(3, 0) {
                        info!("Using OpenGL version {}.", version);
                    } else {
                        return Err(Error::UnsupportedOpenglVersion);
                    }
                },

                Err(err) => { warn!("Using unknown OpenGL version: {}", err); },
            }

            // Load OpenGL functions.
            let mut error = None;

            let gl = gl::Gl::load_with(|name| {
                match glc.get_proc_address(name) {
                    Ok(f) if !f.is_null() => {
                        unsafe {
                            mem::transmute(f)
                        }
                    },

                    _ => {
                        if error.is_none() {
                            error = Some(Error::MissingOpenglFunction(String::from(name)));
                        }

                        unsafe {
                            mem::transmute(1usize)
                        }
                    }
                }
            });

            if let Some(err) = error {
                return Err(err);
            }

            // Initial OpenGL setup.
            unsafe {
                gl.AlphaFunc(gl::GREATER, 0.0);
                gl.ClearColor(0.0, 0.0, 0.0, 0.0);
                gl.Enable(gl::ALPHA_TEST);
                gl.Enable(gl::TEXTURE_2D);
                gl.EnableClientState(gl::VERTEX_ARRAY);
                gl.EnableClientState(gl::COLOR_ARRAY);
                check_gl(&gl)?;
            }

            Ok(gl)
        })??;

        // Construct renderer.
        let mut renderer = Renderer {
            context, gl,
            font: UvFont::empty(),
            font_tex: 0,
            gravity: builder.gravity(),
            scale: builder.scale(),
            buf_size: Vec2::new(0, 0),
            bg_buf: Vec::new(),
            fg_buf: Vec::new(),
            dirty: true,
        };

        renderer.set_font(window, font)?;
        renderer.resize_buf(window.size() / font.glyph_size());

        // Success!
        Ok(renderer)
    }

    fn resize_buf (&mut self, size: Vec2<Coord>) {
        if size == self.buf_size {
            return;
        }

        let cell_size = self.font.glyph_size() * self.scale;
        let uv0 = Vec2::new(0.0, 0.0);
        let black = Rgba::black();
        let transparent = Rgba::new(0, 0, 0, 0);

        let mut bg_buf = Vec::with_capacity((size.x * size.y) as usize);
        let mut fg_buf = Vec::with_capacity((size.x * size.y) as usize);

        for y in 0..size.y {
            for x in 0..size.x {
                if x < self.buf_size.x && y < self.buf_size.y {
                    let old_index = (y * self.buf_size.x + x) as usize;
                    bg_buf.push(self.bg_buf[old_index]);
                    fg_buf.push(self.fg_buf[old_index]);
                } else {
                    let pos = cell_size * (x, y);

                    bg_buf.push([
                        Attrib {
                            vertex: pos,
                            tex_coord: uv0,
                            color: black,
                        },
                        Attrib {
                            vertex: pos + (0, cell_size.y),
                            tex_coord: uv0,
                            color: black,
                        },
                        Attrib {
                            vertex: pos + cell_size,
                            tex_coord: uv0,
                            color: black,
                        },
                        Attrib {
                            vertex: pos + (cell_size.x, 0),
                            tex_coord: uv0,
                            color: black,
                        },
                    ]);

                    fg_buf.push([
                        Attrib {
                            vertex: pos,
                            tex_coord: uv0,
                            color: transparent,
                        },
                        Attrib {
                            vertex: pos + (0, cell_size.y),
                            tex_coord: uv0,
                            color: transparent,
                        },
                        Attrib {
                            vertex: pos + cell_size,
                            tex_coord: uv0,
                            color: transparent,
                        },
                        Attrib {
                            vertex: pos + (cell_size.x, 0),
                            tex_coord: uv0,
                            color: transparent,
                        },
                    ]);
                }
            }
        }

        self.buf_size = size;
        self.bg_buf = bg_buf;
        self.fg_buf = fg_buf;
    }

    fn set_cell (&mut self, pt: Vec2<Coord>, ch: C, fg: Rgb<u8>, bg: Rgb<u8>) {
        if pt.x < 0 || pt.x >= self.buf_size.x || pt.y < 0 || pt.y >= self.buf_size.y {
            return;
        }

        let uv = self.font.get(ch);
        let bg = bg.with_alpha(0xff);
        let fg = fg.with_alpha(0xff);

        let mut bg_cell = &mut self.bg_buf[(pt.y * self.buf_size.x + pt.x) as usize];
        bg_cell[0].color = bg;
        bg_cell[1].color = bg;
        bg_cell[2].color = bg;
        bg_cell[3].color = bg;

        let mut fg_cell = &mut self.fg_buf[(pt.y * self.buf_size.x + pt.x) as usize];
        fg_cell[0].tex_coord = uv.0;
        fg_cell[0].color = fg;
        fg_cell[1].tex_coord = uv.x0y1();
        fg_cell[1].color = fg;
        fg_cell[2].tex_coord = uv.1;
        fg_cell[2].color = fg;
        fg_cell[3].tex_coord = uv.x1y0();
        fg_cell[3].color = fg;
    }

    fn set_font (&mut self, window: &Window, font: &Rc<Font<C>>) -> Result<()> {
        let uv = UvFont::init(font.as_ref())?;
        let image_size = uv.image().size();
        let mut font_tex = 0;

        self.context.with(window, |_| {
            unsafe {
                self.gl.GenTextures(1, &mut font_tex);
                self.gl.BindTexture(gl::TEXTURE_2D, font_tex);

                self.gl.TexImage2D(gl::TEXTURE_2D, 0,
                                   gl::RGB as GLint,
                                   image_size.x, image_size.y,
                                   0,
                                   gl::RGB, gl::UNSIGNED_BYTE,
                                   uv.image().buf().as_ptr() as *const _);

                self.gl.TexParameteri(gl::TEXTURE_2D, gl::TEXTURE_MIN_FILTER, gl::NEAREST as GLint);
                self.gl.TexParameteri(gl::TEXTURE_2D, gl::TEXTURE_MAG_FILTER, gl::NEAREST as GLint);

                if let Err(err) = check_gl(&self.gl) {
                    if font_tex != 0 {
                        self.gl.DeleteTextures(1, &font_tex);
                    }
                    return Err(err);
                }

                if self.font_tex != 0 {
                    self.gl.DeleteTextures(1, &self.font_tex);
                }
            }

            Ok(())
        })??;

        self.font_tex = font_tex;
        self.font = uv;
        self.init_vertices();
        Ok(())
    }

    fn set_scale (&mut self, scale: Vec2<Coord>) {
        self.scale = scale;
        if self.scale.x < 1 { self.scale.x = 1; }
        if self.scale.y < 1 { self.scale.y = 1; }
        self.init_vertices();
    }

    fn update (&mut self, window: &Window) -> Result<()> {
        if !self.dirty {
            return Ok(());
        }

        let window_size = window.size();
        let view_size = self.font.glyph_size() * self.scale * self.buf_size;
        let view_pos = match self.gravity {
            Gravity::TopLeft => Vec2::new(0, 0),
            Gravity::Center => (window_size - view_size) / 2,
        };
        let stride = mem::size_of::<Attrib>() as GLsizei;

        let result: Result<()> = self.context.with(window, |glc| {
            unsafe {
                self.gl.Viewport(0, 0, window_size.x, window_size.y);
                self.gl.Clear(gl::COLOR_BUFFER_BIT);

                self.gl.MatrixMode(gl::PROJECTION);
                self.gl.LoadIdentity();
                self.gl.Ortho(0.0, window_size.x as GLdouble,
                              window_size.y as GLdouble, 0.0,
                              -1.0, 1.0);

                self.gl.MatrixMode(gl::MODELVIEW);
                self.gl.LoadIdentity();
                self.gl.Translatef(view_pos.x as GLfloat, view_pos.y as GLfloat, 0.0);

                self.gl.BindTexture(gl::TEXTURE_2D, 0);
                self.gl.DisableClientState(gl::TEXTURE_COORD_ARRAY);
                self.gl.VertexPointer(2, gl::INT, stride, mem::transmute(&self.bg_buf[0][0].vertex));
                self.gl.ColorPointer(4, gl::UNSIGNED_BYTE, stride, mem::transmute(&self.bg_buf[0][0].color));
                self.gl.DrawArrays(gl::QUADS, 0, self.bg_buf.len() as GLsizei);

                self.gl.BindTexture(gl::TEXTURE_2D, self.font_tex);
                self.gl.EnableClientState(gl::TEXTURE_COORD_ARRAY);
                self.gl.VertexPointer(2, gl::INT, stride, mem::transmute(&self.fg_buf[0][0].vertex));
                self.gl.TexCoordPointer(2, gl::FLOAT, stride, mem::transmute(&self.fg_buf[0][0].tex_coord));
                self.gl.ColorPointer(4, gl::UNSIGNED_BYTE, stride, mem::transmute(&self.fg_buf[0][0].color));
                self.gl.DrawArrays(gl::QUADS, 0, self.fg_buf.len() as GLsizei);

                check_gl(&self.gl)?;
            }

            glc.swap_buffers()?;
            Ok(())
        })?;

        result?;
        self.dirty = false;
        Ok(())
    }
}

/// Check for OpenGL errors.
unsafe fn check_gl (gl: &gl::Gl) -> Result<()> {
    match gl.GetError() {
        0 => Ok(()),
        n => Err(Error::OpenglError(n))
    }
}
