// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::cell::{Cell, RefCell};
use std::cmp;
use std::rc::Rc;

use dcolor::{Color, Rgb};
use dvec::Vec2;
use dwindow::{Display, Event, Window, WindowBuilder};
use dwindow::event::*;
use try_from::TryInto;

use ::Coord;
use error::{Error, Result};
use font::Font;
use render::{Render, Renderer};
use text::{Fallback, FallbackValue, IterChars};

// Constants
const DEFAULT_WINDOW_SIZE: Vec2<Coord> = Vec2 { x: 640, y: 480 };

/// Gravity of the window's content.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Gravity {
    TopLeft,
    Center,
}

/// Action to take when the user resizes the window.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum ResizeMode {
    NoAction,
    ResizeBuffer,
    Scale,
    ScaleFit,
}

/// Creation parameters for `Terminal`.
#[derive(Clone)]
pub struct TerminalBuilder<'a, C> {
    font: Option<Rc<Font<C>>>,
    gravity: Gravity,
    resize_mode: ResizeMode,
    scale: Vec2<Coord>,
    size: Option<Vec2<Coord>>,
    window_builder: WindowBuilder<'a>,
}

impl<'a, C> TerminalBuilder<'a, C>
    where C: Copy + FallbackValue + Ord
{
    pub fn build (&self) -> Result<Terminal<C>> {
        let visual = <Renderer<C> as Render<C>>::choose_visual(self.display())?;
        let mut window_builder = self.window_builder.clone();

        let font = match self.font {
            None => { return Err(Error::NoFont); },
            Some(ref f) => f.clone(),
        };
        let cell_size = font.glyph_size() * self.scale;
        let window_size = match self.size {
            None => DEFAULT_WINDOW_SIZE,
            Some(size) => size * cell_size,
        };

        window_builder = window_builder.with_size(window_size);
        let window = window_builder.build()?;
        let renderer = Renderer::init(self, &window, visual)?;
        let mut buf_size = window.size() / cell_size;

        if buf_size.x < 0 { buf_size.x = 0; }
        if buf_size.y < 0 { buf_size.y = 0; }

        Ok(Terminal {
            window, font,
            renderer: RefCell::new(renderer),
            buf_size: Cell::new(buf_size),
            scale: Cell::new(self.scale),
            gravity: Cell::new(self.gravity),
            resize_mode: self.resize_mode,
            pos: Cell::new(Vec2::new(0, 0)),
            fg: Cell::new(Rgb::white()),
            bg: Cell::new(Rgb::black()),
        })
    }

    pub fn display (&self) -> &'a Display { self.window_builder.display() }
    pub fn font (&self) -> Option<&Rc<Font<C>>> { self.font.as_ref() }
    pub fn gravity (&self) -> Gravity { self.gravity }
    pub fn resize_mode (&self) -> ResizeMode { self.resize_mode }
    pub fn scale (&self) -> Vec2<Coord> { self.scale }
    pub fn size (&self) -> Option<Vec2<Coord>> { self.size }
    pub fn title (&self) -> Option<&str> { self.window_builder.title() }

    #[cfg(feature = "built-in-font")]
    pub fn with_built_in_font (mut self) -> TerminalBuilder<'a, C>
        where C: Ord, ::cp437::Char: ::try_from::TryInto<C>,
    {
        self.font = Some(Rc::new(Font::init_built_in()));
        self
    }

    pub fn with_center_gravity (mut self) -> TerminalBuilder<'a, C> {
        self.gravity = Gravity::Center;
        self
    }

    pub fn with_font<T: Into<Rc<Font<C>>>> (mut self, font: T) -> TerminalBuilder<'a, C> {
        self.font = Some(font.into());
        self
    }

    pub fn with_resize_mode (mut self, mode: ResizeMode) -> TerminalBuilder<'a, C> {
        self.resize_mode = mode;
        self
    }

    pub fn with_scale<T: Into<Vec2<Coord>>> (mut self, scale: T) -> TerminalBuilder<'a, C> {
        self.scale = scale.into();
        if self.scale.x < 1 { self.scale.x = 1; }
        if self.scale.y < 1 { self.scale.y = 1; }
        self
    }

    pub fn with_size<T: Into<Vec2<Coord>>> (mut self, size: T) -> TerminalBuilder<'a, C> {
        let mut size = size.into();
        if size.x < 1 { size.x = 1; }
        if size.y < 1 { size.y = 1; }
        self.size = Some(size);
        self
    }

    pub fn with_title<T: Into<String>> (mut self, title: T) -> TerminalBuilder<'a, C> {
        self.window_builder = self.window_builder.with_title(title);
        self
    }
}

/// Terminal window.
pub struct Terminal<C> {
    window: Window,
    renderer: RefCell<Renderer<C>>,
    font: Rc<Font<C>>,
    buf_size: Cell<Vec2<Coord>>,
    scale: Cell<Vec2<Coord>>,
    gravity: Cell<Gravity>,
    resize_mode: ResizeMode,
    pos: Cell<Vec2<Coord>>,
    fg: Cell<Rgb<u8>>,
    bg: Cell<Rgb<u8>>,
}

impl<C> Terminal<C>
    where C: Copy + FallbackValue + Ord
{
    pub fn clear (&self, color: Rgb<u8>) {
        self.renderer.borrow_mut().clear(color);
    }

    pub fn move_to (&self, pos: Vec2<Coord>) {
        self.pos.set(pos);
    }

    pub fn new (display: &Display) -> TerminalBuilder<C> {
        TerminalBuilder {
            font: None,
            gravity: Gravity::TopLeft,
            resize_mode: ResizeMode::ResizeBuffer,
            scale: Vec2::new(1, 1),
            size: None,
            window_builder: display.new_window(),
        }
    }

    pub fn put_char (&self, ch: C) {
        let pos = self.pos.get();
        self.set_cell(pos, ch, self.fg.get(), self.bg.get());
        self.pos.set(pos + (1, 0));
    }

    pub fn put_chars<T: IterChars<C>> (&self, s: T) {
        let mut pos = self.pos.get();
        let fg = self.fg.get();
        let bg = self.bg.get();
        let mut renderer_ref = self.renderer.borrow_mut();

        for ch in s.iter_chars() {
            renderer_ref.set_cell(pos, ch, fg, bg);
            pos.x += 1;
        }

        self.pos.set(pos);
    }

    pub fn put_str<T: AsRef<str>> (&self, s: T)
        where char: TryInto<C>
    {
        let mut pos = self.pos.get();
        let fg = self.fg.get();
        let bg = self.bg.get();
        let mut renderer_ref = self.renderer.borrow_mut();

        for ch in s.as_ref().chars() {
            renderer_ref.set_cell(pos, ch.try_into().fallback(), fg, bg);
            pos.x += 1;
        }

        self.pos.set(pos);
    }

    pub fn run<F> (&self, display: &Display, mut f: F) -> Result<()>
        where F: FnMut(Event)
    {
        let window_id = self.window.id();
        let mut error = None;

        self.window.set_visible(true)?;

        display.run_passive(|event| {
            if error.is_some() {
                display.stop();
            }

            match event {
                Event::Button(mut ev) => {
                    if ev.sender == window_id {
                        ev.pos = self.pt_to_cell(ev.pos);
                        f(Event::Button(ev));
                    }
                },

                Event::Close(ev) => {
                    if ev.sender == window_id {
                        f(event);
                    }
                },

                Event::Destroy(ev) => {
                    if ev.sender == window_id {
                        f(event);
                        display.stop();
                    }
                },

                Event::Expose(ev) => {
                    if ev.sender == window_id {
                        self.renderer.borrow_mut().damage(ev.rect);
                    }
                },

                Event::Focus(ev) => {
                    if ev.sender == window_id {
                        f(event);
                    }
                },

                Event::Key(mut ev) => {
                    if ev.sender == window_id {
                        ev.pos = self.pt_to_cell(ev.pos);
                        f(Event::Key(ev));
                    }
                },

                Event::Motion(mut ev) => {
                    if ev.sender == window_id {
                        ev.pos = self.pt_to_cell(ev.pos);
                        f(Event::Motion(ev));
                    }
                },

                Event::Resize(ev) => {
                    if ev.sender == window_id {
                        match self.resize_mode {
                            ResizeMode::NoAction => (),

                            ResizeMode::ResizeBuffer => {
                                let cell_size = self.font.glyph_size() * self.scale.get();
                                let mut buf_size = ev.size / cell_size;
                                if buf_size.x < 0 { buf_size.x = 0; }
                                if buf_size.y < 0 { buf_size.y = 0; }
                                self.renderer.borrow_mut().resize_buf(buf_size);
                                self.buf_size.set(buf_size);
                            },

                            ResizeMode::Scale => {
                                let glyph_size = self.font.glyph_size();
                                let mut scale = ev.size / (self.buf_size.get() * glyph_size);
                                let min_scale = cmp::max(cmp::min(scale.x, scale.y), 1);
                                scale = Vec2::new(min_scale, min_scale);
                                self.renderer.borrow_mut().set_scale(scale);
                                self.scale.set(scale);
                            },

                            ResizeMode::ScaleFit => {
                                let glyph_size = self.font.glyph_size();
                                let mut scale = ev.size / (self.buf_size.get() * glyph_size);
                                if scale.x < 1 { scale.x = 1; }
                                if scale.y < 1 { scale.y = 1; }
                                self.renderer.borrow_mut().set_scale(scale);
                                self.scale.set(scale);
                            },
                        }

                        let cell_size = self.font.glyph_size() * self.scale.get();
                        let buf_size = ev.size / cell_size;

                        f(Event::Resize(ResizeEvent {
                            sender: ev.sender,
                            size: buf_size,
                        }));
                    }
                },

                Event::Signal(_) => { f(event); },

                Event::Update(_) => {
                    if let Err(err) = self.renderer.borrow_mut().update(&self.window) {
                        error = Some(err);
                    }
                    f(event);
                },

                Event::Wheel(mut ev) => {
                    if ev.sender == window_id {
                        ev.pos = self.pt_to_cell(ev.pos);
                        f(Event::Wheel(ev));
                    }
                },
            }
        })?;

        match error {
            None => Ok(()),
            Some(err) => Err(err),
        }
    }

    pub fn set_bg (&self, bg: Rgb<u8>) {
        self.bg.set(bg);
    }

    pub fn set_cell (&self, pt: Vec2<Coord>, ch: C, fg: Rgb<u8>, bg: Rgb<u8>) {
        self.renderer.borrow_mut().set_cell(pt, ch, fg, bg);
    }

    pub fn set_fg (&self, fg: Rgb<u8>) {
        self.fg.set(fg);
    }

    pub fn size (&self) -> Vec2<Coord> { self.buf_size.get() }
}

impl<C> Terminal<C>
    where C: Copy + FallbackValue + Ord
{
    fn pt_to_cell (&self, pt: Vec2<Coord>) -> Vec2<Coord> {
        let cell_size = self.font.glyph_size() * self.scale.get();
        let origin = match self.gravity.get() {
            Gravity::TopLeft => Vec2::new(0, 0),
            Gravity::Center => {
                let view_size = self.buf_size.get() * cell_size;
                (self.window.size() - view_size) / 2
            },
        };
        (pt - origin) / cell_size
    }
}
