// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::iter::Cloned;
use std::slice;
use std::str;
use std::vec;

/// Uses a fallback value if some operation failes.
pub trait Fallback {
    type Value;
    fn fallback (self) -> Self::Value;
}

impl<T: FallbackValue> Fallback for Option<T> {
    type Value = T;

    fn fallback (self) -> T {
        match self {
            None => T::fallback_value(),
            Some(t) => t,
        }
    }
}

impl<T: FallbackValue, E> Fallback for Result<T, E> {
    type Value = T;

    fn fallback (self) -> T {
        match self {
            Ok(t) => t,
            Err(_) => T::fallback_value(),
        }
    }
}

/// Fallback value if some operation fails.
pub trait FallbackValue {
    fn fallback_value () -> Self;
}

impl FallbackValue for char {
    fn fallback_value () -> char { '\u{fffd}' }
}

/// Iterate characters in a string.
pub trait IterChars<T> {
    type Iter : Iterator<Item = T>;
    fn iter_chars (self) -> Self::Iter;
}

impl<'a, T: 'a + Copy> IterChars<T> for &'a [T] {
    type Iter = Cloned<slice::Iter<'a, T>>;
    fn iter_chars (self) -> Cloned<slice::Iter<'a, T>> { self.iter().cloned() }
}

impl<'a> IterChars<char> for &'a str {
    type Iter = str::Chars<'a>;
    fn iter_chars (self) -> str::Chars<'a> { self.chars() }
}

impl<'a, T: 'a + Copy> IterChars<T> for &'a Vec<T> {
    type Iter = Cloned<slice::Iter<'a, T>>;
    fn iter_chars (self) -> Cloned<slice::Iter<'a, T>> { self.iter().cloned() }
}

impl<T> IterChars<T> for Vec<T> {
    type Iter = vec::IntoIter<T>;
    fn iter_chars (self) -> vec::IntoIter<T> { self.into_iter() }
}

impl<'a> IterChars<char> for &'a String {
    type Iter = str::Chars<'a>;
    fn iter_chars (self) -> str::Chars<'a> { self.chars() }
}
